/*===========================================================================*/
/* Driver routines for CCD controller digital signal processor interface     */
/*===========================================================================*/
#define DSPDVR_C

#include <stdio.h>
#include <stdlib.h>                 /* swab() */
#include <unistd.h>                 /* close(),ioctl(),sleep() */
#include <sys/types.h>
#include <fcntl.h>
#include <time.h>                   /* nanosleep() */
#include <sys/time.h>               /* gettimeofday() */

#ifdef LINUX
#include <sys/ioctl.h>              /* ioctl() */
#include <linux/ioctl.h>
#include <linux/types.h>
#include "amcc.h"
extern void swab(const void*,void*,size_t);
#endif

#include "dspdvr.h"

/*========== definitions ====================================================*/
/* Define the DSP address locations for the following control parameters */
#define  CMD            0x0000      /* Command flags */
#define  OPFLAGS        0x0001      /* Operational flags */
#define  NROWS          0x0002      /* Number of rows */
#define  NCOLS          0x0003      /* Number of columns */
#define  NFT            0x0004      /* Number of rows for frame transfer */
#define  NFLUSH         0x0005      /* Number of rows to flush */
#define  NCCD           0x0006      /* Number of output channels */
#define  NAMPS          0x0006      /* Number of amps or detectors */
#define  NCH            0x0006      /* Number of output channels */
#define  NPIX           0x0007      /* (not used) */
#define  VBIN           0x0008      /* Vertical (parallel) binning */
#define  HBIN           0x0009      /* Horizontal (serial) binning */
#define  VSKIP          0x000A      /* V prescan or offset (rows)    */
#define  HSKIP          0x000B      /* H prescan or offset (columns) */
#define  VSUB           0x000C      /* V subraster size */
#define  HSUB           0x000D      /* H subraster size */
#define  NEXP           0x000E      /* Number of exposures */
#define  NSHUFFLE       0x000F      /* Number of shuffle-rows */

#define  EXP_TIME       0x0010      /* CCD exposure time (millisec) */
#define  TEMP           0x0011      /* Temperature sensor */
#define  GAIN           0x0012      /* Sig_proc gain */
#define  USEC           0x0013      /* Sig_proc sample time */
#define  OPCH           0x0014      /* Output channel */
#define  HDIR           0x0015      /* Horizontal clk direction */
#define  HCLKS          0x0015      /* Horizontal clk direction */
#define  LINK           0x0016      /* Link type 0=wire 1=fiber */

#ifdef GCAM
#define  VBIAS          0x0018      /* CCD bias voltages */
#define  VCLK           0x0020      /* CCD clock voltages */
#endif // GCAM

#define  ETIMER         0x0018      /* Exposure countdown timer */
#define  ESTATE         0x0019      /* Shutter open/closed */
#define  TEC            0x001A      /* TEC setting */
#define  PID            0x001B      /* Temp control PID loop flag*/

#define  HEAT           0x0100      /* Heater resistor voltages */
#define  TCCD           0x0108      /* CCD temp sensor reading(s) */
#define  TSET           0x0110      /* CCD temp set-points */

#ifdef BASE
#define  VBIAS          0x0120      /* CCD bias voltages */
#define  VCLK           0x0180      /* CCD clock voltages */
#endif // BASE

#ifdef H2RG
#define  VBIAS          0x0120      /* H2RG bias voltages */
#define  VCLK           0x0180      /* H2RG clock voltages (not used) */
#endif // H2RG

#define  MAX_RETRYS     100         /* max consecutive read retrys */
#define  MAX_REWRITE    50          /* max consecutive write retrys */
#define  TIMEOUT        5           /* read time-out (msec) */
#define  UTIMEOUT       1000        /* read time-out (usec) */
#define  RD_TIMEOUT     5           /* read time-out (msec) */
#define  PLL_SLEEP      1000        /* Sleep time at start-up (msec) */

#define  LO_G           (5.0E-3)    /* Low  gain DAC 5mV/DN */
#define  HI_G           (10.0E-3)   /* High gain DAC  10 mV/DN */
#define  DZ             4096        /* DAC zero voltage set-point */
#define  CLK_SPEED      80          /* DSP clock speed in MHz */
#define  FIFO_SIZE      16384       /* FIFO is 16384 words */

#define  NITEMS         2048        /* simulated memory depth */

/*=============== function prototypes =======================================*/
void msleep(int msec);
void udelay(int u_secs);

/*========== local variables ================================================*/
static int dsp;                 /* Assigned device number */
static int dsp_link;            /* 0=no link, 1= serial link working */
static int reg, xmem[2048], ymem[2048], pmem[4096];
static int BIGxENDIAN;
static int debug=1;

/*========== Open/configure/close the serial link =====================*/
int  
dsp_open(void)
{
    extern int dsp;

    union                                        /* determine endian */
    {
        long l;
        char c[sizeof(long)];
    } u;
    u.l = 1;

    if (u.c[sizeof(long)-1] == 1)
        BIGxENDIAN = 1;
    else
        BIGxENDIAN = 0;

    dsp = open("/dev/amcc0", O_RDWR);
    if (dsp < 0) {                              /* DSP offline */
        printf("DSP offline\n");
        printf("Error opening /dev/amcc0 \n");
    }
    else {
        printf("Device /dev/amcc0 opened \n");
    }
    return(dsp);
}

void
dsp_close(void)
{
    extern int dsp;

    if (dsp > 0) close(dsp);
    printf("Device /dev/amcc0 closed\n");
}

void
dsp_reset(void)
{
    extern int dsp;
    int j;

    /* Reset the DSP through the PCI interface, wait for PLL to lock */
    printf("Resetting the DSP\n");
    printf("Waiting for PLL to lock\n");
    if (dsp > 0) (void)ioctl(dsp, AMCC_RESET_ADDON);

    /* Set up the dsp simulator, if required */
    if (dsp < 0){                           /* DSP offline */
        reg = 0;
        for (j=0; j<NITEMS; j++){
            xmem[j]=0;
            ymem[j]=0;
        }
        printf("Clearing dsp simulator memory\n");
    }
    msleep(PLL_SLEEP);
}

int
dsp_clear_fifo(void)
{
    extern int dsp;
    int r, icsr, count=0, retrys=0;
    u_short rbuf[1];

    /* Clean out any stray words in read FIFO.  On /RESET, the DSP   */
    /* has the nasty habit of sending some stray bits out of the SSI */

    if (dsp < 0) return(0);                     /* DSP offline */

    (void)ioctl(dsp, AMCC_RESET_FIFO, &count);
    if (count > 0)printf("Clearing FIFO  count=  %d\n",count);
    return(count);
}


/* ------------------------------------------------------------------------- */
/* utility functions                                                         */
/* ------------------------------------------------------------------------- */
void 
msleep(int msec)
{
    struct timespec delay;
    struct timespec rem;

    if (msec < 0) return;

    delay.tv_sec  = msec/1000;
    delay.tv_nsec = (msec % 1000) * 1000000;

    while (nanosleep(&delay,&rem)<0){       /* prevents early return */
        delay.tv_sec = rem.tv_sec;
        delay.tv_nsec = rem.tv_nsec;
    }
}

void 
udelay(int u_sec)
{
    struct timeval clockA, clockB;
    u_long timeA, timeB;

    if (u_sec > 50000) {                         /* that's enough :-) */
        msleep(u_sec/1000); 
        return;
    }
    gettimeofday(&clockB,NULL);                         /* start time */
    timeA = (u_long)clockB.tv_usec;
    timeB = (u_long)clockB.tv_usec + (u_long)u_sec;     /* end time */

    while (timeA < timeB){
        gettimeofday(&clockA,NULL);
        timeA = (u_long)(clockA.tv_sec-clockB.tv_sec)*1000000 +
                (u_long)clockA.tv_usec;
    }
}

/*========== Set up and load the S-Records into the DSP =====================*/
int  
dsp_init(char *srec_file)
{
    extern int dsp, dsp_link, debug;
    int srec_count, srec_sent, retrys, rewrite;
    int i,j,r, rc, wc, scan_index, bc;
    int checksum, nwords;
    int num, addr, val, mem_space;
    u_short wbuf[256], rbuf[256];
    char srec[525], cmd[4];
    FILE *fp;
    u_char *buf = (u_char *)wbuf;

    /* Reset the DSP, wait for PLL lock and open the s-record file */
    /* Clear out any stray words in the read FIFO from DSP reset   */
    dsp_reset();
    dsp_clear_fifo();
    dsp_link = 0;

    srec_count = 0;
    srec_sent = 0;

    if( (fp = fopen( srec_file, "r")) != NULL) {
        printf("Loading s-records from file: %s\n", srec_file);
    }
    else {
        printf("Could not open file: %s\n", srec_file);
        return(0);
    }

    while(fgets(srec,sizeof(srec)-1,fp)!=NULL){
        if (srec[0] != 'S') continue;
        srec_count++;
        sscanf(srec,"%2s %2x",cmd,&num);
        scan_index=4;
        nwords = (num-3)/3;

        /* Scan the two byte (S0 or S1) or three byte (S2) address field */
        switch(srec[1]) {
        case '0':
            sscanf(&srec[scan_index],"%4x",&addr);
            scan_index+=4;
            mem_space = (addr&0xff);
            break;
        case '1':
            sscanf(&srec[scan_index],"%4x",&addr);
            scan_index+=4;
            break;
        case '2':
            sscanf(&srec[scan_index],"%6x",&addr);
            scan_index+=6;
            break;
        case '8':
            sscanf(&srec[scan_index],"%6x",&addr);
            scan_index+=6;
            mem_space = 0x08;
            break;
        case '9':
            sscanf(&srec[scan_index],"%4x",&addr);
            mem_space = 0x08;
            scan_index+=4;
            break;
        default:
            break;
        }

/*   buf[0] = number of words, excluding first two header words            *
 *      [1] = dummy byte, coded as zero                                    *
 *      [2] = memory space code X=1, Y=2, P=4, END=8                       *
 *      [ ]                                                                *
 *      [0] = start address (bits 8..0)                                    *
 *      [1] = start address (bits 15..8)                                   *
 *      [2] = start address (bits 23..16)                                  *
 *      [ ]                                                                */

        /* Arrange the outgoing bytes in the write buffer */
        bc = 0;                                     /* byte counter */
        buf[bc++] = (u_char)(nwords>>0 & 0xff);     /* number of 24-bit words*/
        buf[bc++] = (u_char)(nwords>>8 & 0xff);
        buf[bc++] = (u_char)(mem_space & 0xff);     /* memory space */
        buf[bc++] = 0;

        buf[bc++] = (u_char)(addr>>0 & 0xff);       /* address bits  7..0 */
        buf[bc++] = (u_char)(addr>>8 & 0xff);       /* address bits 15..8 */
        buf[bc++] = (u_char)(addr>>16 & 0xff);      /* address bits 23..16 */
        buf[bc++]=0;

        while (scan_index<=2*num){
            sscanf(&srec[scan_index],"%6x",&val);
            buf[bc++] = (val>>0 & 0xff);
            buf[bc++] = (val>>8 & 0xff);
            buf[bc++] = (val>>16 & 0xff);
            buf[bc++]=0;
            if (mem_space==1)xmem[addr++]=(val & 0xffffff);
            if (mem_space==2)ymem[addr++]=(val & 0xffffff);
            scan_index+=6;
        }

        /* Calculate checksum */
        checksum = 0;
        for (j=0; j<bc/sizeof(u_short); j++) checksum += wbuf[j];
        checksum &= 0xffff;

        if (dsp < 0) continue;                  /* DSP offline */

        /*  Write s-record to the dsp, and read back checksum */
        if (srec[1]=='1'||srec[1]=='2') {
            rbuf[0] = ~checksum;
            rewrite = 0;
            while (rbuf[0] != checksum && rewrite<MAX_REWRITE){
                dsp_clear_fifo();
                wc = write(dsp,wbuf,bc);
                rewrite++;
                retrys = 0;
                while (retrys<MAX_RETRYS){
                    r = read(dsp,rbuf,sizeof(u_short));
                    if (r > 0) break;
                    udelay(UTIMEOUT);
                    retrys++;
                }
                if(retrys>0 && debug>0){ 
                    printf("srec=%3d  r=%d  retrys=%3d rbuf=%04x check=%04x\n",
                        srec_count, r, retrys, rbuf[0], checksum);
                }
            }
            srec_sent++;

            if(rewrite >= MAX_REWRITE) {
                printf("Serial link comm retry failure\n");
                dsp_link=0;
                break;
            }
            else {
                dsp_link=1;
            }
        }
        /* write last record -- no checksum on this one, then   *
         * wait for dsp to start executing loaded program       */
        if (srec[1] == '8' || srec[1] == '9'){
            wc = write(dsp,wbuf,bc);
            srec_sent++;
            msleep(TIMEOUT);        
        }
    }
    fclose(fp);

    /* Clean up after the DSP as it restarts its peripherals */
    dsp_clear_fifo();
    printf("S-records processed  %3d\n",srec_count);
    printf("S-records uplinked   %3d\n",srec_sent);
    return(dsp_link);
}

int
dsp_initII(char *srec_file)
{
    extern int dsp, dsp_link;
    int srec_count, srec_sent, retrys, rewrite;
    int i,j,m,r=0, rc, wc, scan_index, bc;
    int nwords, num, addr, val, mem_space;
    u_short wbuf[256], rbuf[256];
    char srec[525], cmd[4];
    FILE  *fp;
    const int verbose=0;

    dsp_reset();
    dsp_clear_fifo();
    dsp_link = 0;

    srec_count = 0;
    srec_sent = 0;

    if( (fp = fopen( srec_file, "r")) != NULL) {
        printf("Loading s-records from file: %s\n", srec_file);
    }
    else {
        printf("Could not open file: %s\n", srec_file);
        return(0);
    }

    if (dsp < 0)dsp_link = 1;               // DSP offline

    while(fgets(srec,sizeof(srec)-1,fp)!=NULL){
        if (srec[0] != 'S') continue;
        srec_count++;
        sscanf(srec,"%2s %2x",cmd,&num);
        scan_index=4;
        nwords = (num-3)/3;

        /* Scan the two byte (S0 or S1) or three byte (S2) address field */
        switch(srec[1]) {
        case '0':
            sscanf(&srec[scan_index],"%4x",&addr);
            scan_index+=4;
            mem_space = (addr&0xff);
            break;
        case '1':
            sscanf(&srec[scan_index],"%4x",&addr);
            scan_index+=4;
            break;
        case '2':
            sscanf(&srec[scan_index],"%6x",&addr);
            scan_index+=6;
            break;
        case '8':
            sscanf(&srec[scan_index],"%6x",&addr);
            scan_index+=6;
            mem_space = 0x08;
            break;
        case '9':
            sscanf(&srec[scan_index],"%4x",&addr);
            mem_space = 0x08;
            scan_index+=4;
            break;
        default:
            break;
        }

/*      wbuf[0] = command word | complement                         *
 *          [1] = number of 16-bit data words                       *
 *          [2] = memory space code X=1, Y=2, P=4, END=8            *
 *          [3] = start address (bits 15..0)                        *
 *          [ ] = data words                                        */

        bc = 0;
        wbuf[bc++] = 0xfe01;
        wbuf[bc++] = nwords & 0xffff;
        wbuf[bc++] = mem_space & 0xffff;
        wbuf[bc++] = addr & 0xffff;

        while (scan_index<=2*num){
            sscanf(&srec[scan_index],"%6x",&val);
            wbuf[bc++] = (val>>0 & 0xffff);
            wbuf[bc++] = (val>>16 & 0xff);
            if (mem_space==1)xmem[addr++]=(val & 0xffffff);
            if (mem_space==2)ymem[addr++]=(val & 0xffffff);
            scan_index+=6;
        }
        bc = 2*bc;      // convert to bytes

        if (dsp < 0) continue;                      // DSP offline

        if (srec[1]=='1'||srec[1]=='2'||srec[1]=='8'||srec[1]=='9') {
            rewrite=0;
            rbuf[0]=~wbuf[0];
            while (rbuf[0]!=wbuf[0] && rewrite<MAX_REWRITE){
                dsp_clear_fifo();
                wc=write(dsp,wbuf,bc);
                rewrite++;
                retrys=0;
                while (retrys<MAX_RETRYS){
                    r=read(dsp,rbuf,bc);
                    if (r==bc) break;
                    retrys++;
                    udelay(UTIMEOUT); 
                }
            }

            srec_sent++;
            if(rewrite >= MAX_REWRITE) {
                printf("Serial link comm retry failure\n");
                dsp_link=0;
                break;
            }
            else {
                dsp_link = 1;
            }
        }
    }
    fclose(fp);

    /* Wait for dsp to start executing loaded program */
    msleep(TIMEOUT);

    /* Clean up after the DSP as it restarts its peripherals */
    dsp_clear_fifo();
    printf("S-records processed    %3d\n",srec_count);
    printf("S-records uplinked     %3d\n",srec_sent);
    return(dsp_link);
}

/*==========  Utility routines used by various subprograms ==================*/
int
dsp_cmd(int cmd)
{
    extern int dsp_link;
    int r, ncmd, val=0, retrys=0;

    if (dsp_link==0)return(0);

#ifdef BASE
    while (dsp_clear_fifo()){
        if (retrys++ > MAX_RETRYS)return(-1);
        (void)dsp_write16(DSP_DUMMY);
        (void)dsp_read16(&val);
    }
#endif  // BASE
    ncmd = (cmd | (~cmd & 0xFF)<<8);
    r = dsp_write16(ncmd);
    if (r != sizeof(u_short))return(-1);

#ifdef BASE
    r = dsp_read16(&val);
    if (r!=sizeof(u_short) || val!=ncmd){
        return(-1);
    }
#endif  // BASE
    return(0);
}


/* Send a 16-bit word to the DSP */
int
dsp_write16(int value)
{
    extern int dsp;
    int r=0, retrys=0;
    u_short wbuf[1];

    if (dsp < 0) {                              /* DSP offline */
        reg = (value & 0xffff);
        return(sizeof(u_short));
    }

    wbuf[0] =  (value>>0 & 0xffff);
    if (BIGxENDIAN)swab(wbuf,wbuf,sizeof(u_short));

    while( r<=0 && retrys++ < MAX_RETRYS){
        r = write(dsp, wbuf, sizeof(u_short));
    }
    return(r);
}

/* Receive a 16-bit word from the DSP */
int
dsp_read16(int *value)
{
    extern int dsp;
    int r=0, retrys=0;
    u_short  rbuf[1];

    if (dsp < 0) {                              /* DSP offline */
        *value = (reg & 0xffff);
        return(sizeof(u_short));
    }

    rbuf[0]=0;
    while(r<=0 && retrys < 10*MAX_RETRYS) {
        r = read(dsp,rbuf,sizeof(u_short));
        if (r<=0) {
            retrys++;
            udelay(UTIMEOUT);
        }
    }

    if (BIGxENDIAN)swab(rbuf,rbuf,sizeof(u_short));
    *value = (int)(rbuf[0] & 0xffff);

    return(r);
}

/* Send a 32-bit word to the DSP, 24 LSBs valid */
int
dsp_write24(int value)
{
    extern int dsp;
    int r=0, retrys=0;
    u_short  wbuf[2];

    if (dsp < 0) {                              /* DSP offline */
        reg = (value & 0xffffff);
        return(sizeof(long));
    }

    wbuf[0] = (u_short)(value>>0 & 0xffff);
    wbuf[1] = (u_short)(value>>16 & 0x00ff);
    if (BIGxENDIAN)swab(wbuf,wbuf,sizeof(long));

    while( r<=0 && retrys++ < MAX_RETRYS){
        r = write(dsp, wbuf, sizeof(long));
    }
    return(r);
}

/* Receive a 32-bit word from the DSP, 24 LSBs valid */
int
dsp_read24(int *value)
{
    extern int dsp;
    int r=0, m=0, retrys=0;
    u_short  rbuf[2];
    u_char  *buf = (u_char *)rbuf;

    if (dsp < 0) {                              /* DSP offline */
        *value = (reg & 0xffffff);
        return(sizeof(long));
    }

    while(m < sizeof(long) && retrys < 10*MAX_RETRYS) {
        r = read(dsp,&buf[m],sizeof(long)-m);
        m+=r;
        if (r == 0) {
            retrys += 1;
            udelay(UTIMEOUT);
        }
    }

    if (BIGxENDIAN)swab(rbuf,rbuf,sizeof(long));
    *value = (int) ((rbuf[0] & 0xffff) | ((rbuf[1] & 0x00ff)<<16));
    return(m);
}

/*========== DSP action routines =============================================*/
int
dsp_echo(int sval, int *rval)
{
    int r;

    r = dsp_cmd(DSP_ECHO);
    if (r != 0) return(-1);

    (void)dsp_write16(sval);
    udelay(UTIMEOUT);
    (void)dsp_read16(rval);
    printf("send=%04x  get=%06x \n",sval,*rval);

    return(r);
}

int  
dsp_put(int item, int value)
{
    extern int dsp, xmem[];
    int r=0;

    if (dsp < 0) {                              /* DSP offline */
        if (item < NITEMS) xmem[item] = value;
//      printf("dsp_put  addr= %d   val= %d\n", item, value);
        return(0);
    }

    r = dsp_cmd(DSP_PUT);
    if (r != 0) return(-1);

    r = dsp_write24(item);
    if (r != sizeof(long))return(-1);

    r = dsp_write24(value);
    if (r != sizeof(long))return(-1);

    return(0);
}


int
dsp_get(int item, int *value)
{
    extern int dsp, xmem[];
    int r=0;

    if (dsp < 0) {                              /* DSP offline */
        if (item < NITEMS) *value = xmem[item];
//      printf("dsp_get  addr= %d   val= %d\n", item, *value);
        return(0);
    }

    r = dsp_cmd(DSP_GET);
    if (r != 0) return(-1);

    r = dsp_write24(item);
    if (r != sizeof(long))return(-1);

    r = dsp_read24(value);
    if (r != sizeof(long))return(-1);

    return(0);
}

int
dsp_abort(void)
{
    int r;

    r = dsp_cmd(DSP_ABORT);
    return(r);
}

int
set_seqmode(int mode)
{
    int r=0;

    r = dsp_put(OPFLAGS,mode);
    return(r);
}

int  
set_shutter(int arg)
{
    int r;

    if (arg == 0){
        r = dsp_cmd(BEAM_OFF);
    }
    else {
        r = dsp_cmd(BEAM_ON);
    }
    return(r);
}

int 
exp_start(void)
{
    int r;

    r = dsp_cmd(EXPCCD);
    return(r);
}

int
exp_pause(void)
{
    int r;

    r = dsp_cmd(EXP_PAUSE);
    return(r);
}

int
exp_resume(void)
{
    int r;

    r = dsp_cmd(EXP_RESUME);
    return(r);
}

int
exp_status(int *estate, int *etimer)
{
/* ESTATE  0x0=shutter closed  0x1=shutter open */
/* ETIMER  Exposure countdown timer (milliseconds) */

    int r=0;

    dsp_get(ESTATE,estate);
    dsp_get(ETIMER,etimer);

    return(r);
}

int
read_ccd(void)
{
    int r;

    printf("Starting  read_ccd\n");
    r = dsp_cmd(READCCD);
    return(r);
}

int 
read_line(void)
{
    int r;

    r = dsp_cmd(READLINE);
    return(r);
}

int
get_ccddata(u_short pixbuf[], int npix, int *cnt)
{
    extern int dsp;
    int r=0, m=0, retrys=0, nbytes;
    u_char *rbuf = (u_char *)pixbuf;

    nbytes = npix*sizeof(u_short);

    if (dsp < 0){                               /* DSP offline */
//        for (r=0; r < npix; r+=2){
//            pixbuf[r+0] = 500 + (int)(100.0*rand()/(RAND_MAX+1.0));
//            pixbuf[r+1] = 550 + (int)(100.0*rand()/(RAND_MAX+1.0));
//        }

        for (r=0; r < npix; r++){
            pixbuf[r] = 25000 + 20*r;
        }

#ifdef GCAM_TEST
        for (r=0; r < npix; r+=4){
            pixbuf[r+0] = 500 + (int)(100.0*rand()/(RAND_MAX+1.0));
            pixbuf[r+1] = 550 + (int)(100.0*rand()/(RAND_MAX+1.0));
            pixbuf[r+2] = 600 + (int)(100.0*rand()/(RAND_MAX+1.0));
            pixbuf[r+3] = 650 + (int)(100.0*rand()/(RAND_MAX+1.0));
        }
#endif // GCAM_TEST
        msleep(5);
        return(nbytes);
    }

    while (m < nbytes  && retrys < MAX_RETRYS) {
        r = read(dsp, &rbuf[m], nbytes-m);
        m += r;

        if (r > 0){
            retrys = 0;
        }
        else {
            *cnt += 1;
            retrys += 1;
            msleep(TIMEOUT);     /* Timeout is now in driver */           
        }
    }

    if (retrys >= MAX_RETRYS)
        printf("Get_ccddata failure %d bytes left\n",nbytes-m);

    if (BIGxENDIAN) swab(pixbuf,pixbuf,m);
    return(m);      /* return number of bytes read */
}

int  
flush_ccd(void)
{
    int r=0;

    /* Send FLUSH command to DSP via SSI or SCI */
    r = dsp_cmd(FLUSH);
    return(r);
}

int
flush_line(void)
{
    int r;

    r = dsp_cmd(FLUSHLINE);
    return(r);
}

int 
shuffle(int dir)
{
    int r;

    if (dir==0){
        r = dsp_cmd(SHUFFLE_F);
    }
    else {
        r = dsp_cmd(SHUFFLE_R);
    }
    return(r);
}

int
set_idlewipe(void)
{
    int r;

    r = dsp_cmd(IDLE_WIPE);
    return(r);
}

/*==========  Routines to change/read the CCD control parameters ============*/
int
set_nrow(int nrow)
{
    int r, val=0;
    r = dsp_put(NROWS, nrow);
    if (r == 0) dsp_get(NROWS, &val);
    return(val);
}

int 
get_nrow(void)
{
    int r, val=0;

    dsp_get(NROWS, &val);
    return(val);
}

int  
set_ncol(int ncol)
{
    int r, val=0;
    r = dsp_put(NCOLS, ncol);
    if (r == 0) dsp_get(NCOLS, &val);
    return(val);
}

int 
get_ncol(void)
{
    int r, val=0;

    dsp_get(NCOLS, &val);
    return(val);
}

int
set_nft(int nft)
{
    int r, val=0;
    r = dsp_put(NFT, nft);
    if (r == 0) dsp_get(NFT, &val);
    return(val);
}

int 
get_nft(void)
{
    int r, val=0;

    dsp_get(NFT, &val);
    return(val);
}

int
set_nflush(int nflush)
{
    int r, val=0;
    r = dsp_put(NFLUSH, nflush);
    if (r == 0) dsp_get(NFLUSH, &val);
    return(val);
}

int 
get_nflush(void)
{
    int r, val=0;

    dsp_get(NFLUSH, &val);
    return(val);
}

int
set_nccd(int nccd)
{
    int r, val=0;
    r = dsp_put(NCCD, nccd);
    if (r == 0) dsp_get(NCCD, &val);
    return(val);
}

int 
get_nccd(void)
{
    int r, val=0;

    dsp_get(NCCD, &val);
    return(val);
}

int
set_namps(int namps)
{
    int r, val=0;
    r = dsp_put(NAMPS, namps);
    if (r == 0) dsp_get(NAMPS, &val);
    return(val);
}

int 
get_namps(void)
{
    int r, val=0;

    dsp_get(NAMPS, &val);
    return(val);
}

int
set_nch(int nch)
{
    int r, val=0;
    r = dsp_put(NCH, nch);
    if (r == 0) dsp_get(NCH, &val);
    return(val);
}

int 
get_nch(void)
{
    int r, val=0;

    dsp_get(NCH, &val);
    return(val);
}

int
set_npix(int npix)
{
    int r, val=0;
    r = dsp_put(NPIX, npix);
    if (r == 0) dsp_get(NPIX, &val);
    return(val);
}

int 
get_npix(void)
{
    int r, val=0;

    dsp_get(NPIX, &val);
    return(val);
}

int
set_vbin(int vbin)
{
    int r, val=0;
    r = dsp_put(VBIN, vbin);
    if (r == 0) dsp_get(VBIN, &val);
    return(val);
}

int 
get_vbin(void)
{
    int r, val=0;

    dsp_get(VBIN, &val);
    return(val);
}

int
set_hbin(int hbin)
{
    int r, val=0;
    r = dsp_put(HBIN, hbin);
    if (r == 0) dsp_get(HBIN, &val);
    return(val);
}

int 
get_hbin(void)
{
    int r, val=0;

    dsp_get(HBIN, &val);
    return(val);
}

int
set_vskip(int vskip)
{
    int r, val=0;
    r = dsp_put(VSKIP, vskip);
    if (r == 0) dsp_get(VSKIP, &val);
    return(val);
}

int 
get_vskip(void)
{
    int r, val=0;

    dsp_get(VSKIP, &val);
    return(val);
}

int
set_hskip(int hskip)
{
    int r, val=0;
    r = dsp_put(HSKIP, hskip);
    if (r == 0) dsp_get(HSKIP, &val);
    return(val);
}

int 
get_hskip(void)
{
    int r, val=0;

    dsp_get(HSKIP, &val);
    return(val);
}

int
set_vsub(int vsub)
{
    int r, val=0;
    r = dsp_put(VSUB, vsub);
    if (r == 0) dsp_get(VSUB, &val);
    return(val);
}

int 
get_vsub(void)
{
    int r, val=0;

    dsp_get(VSUB, &val);
    return(val);
}

int
set_hsub(int hsub)
{
    int r, val=0;
    r = dsp_put(HSUB, hsub);
    if (r == 0) dsp_get(HSUB, &val);
    return(val);
}

int 
get_hsub(void)
{
    int r, val=0;

    dsp_get(HSUB, &val);
    return(val);
}

int
set_nshuffle(int nshuffle)
{
    int r, val=0;
    r = dsp_put(NSHUFFLE, nshuffle);
    if (r == 0) dsp_get(NSHUFFLE, &val);
    return(val);
}

int 
get_nshuffle(void)
{
    int r, val=0;

    dsp_get(NSHUFFLE, &val);
    return(val);
}

int
set_seqtime(int exp_time)
{
    int r, val=0;
    r = dsp_put(EXP_TIME, exp_time);
    if (r == 0) dsp_get(EXP_TIME, &val);
    return(val);
}

int 
get_seqtime(void)
{
    int r, val=0;

    dsp_get(EXP_TIME, &val);
    return(val);
}

int 
set_gain(int gain)
{
    int r, val=0;

    if (gain==0 || gain==1){
        r = dsp_put(GAIN, gain);
        if (r == 0) dsp_get(GAIN, &val);
        r = dsp_cmd(HH_OTHER);
    }
    return(val);
}

int 
get_gain(void)
{
    int r, val=0;

    dsp_get(GAIN, &val);
    return(val);
}

int 
set_usec(int usec)
{
    int r, val=0;

    r = dsp_put(USEC, usec);

    if (r == 0) dsp_get(USEC, &val);
    r = dsp_cmd(HH_OTHER);

    return(val);
}

int 
get_usec(void)
{
    int r, val=0;

    dsp_get(USEC, &val);
    return(val);
}

int 
set_opch(int opch)
{
    int r, val=0;

    if (opch>=0 && opch<=7){
        r = dsp_put(OPCH, opch);
        if (r == 0)dsp_get(OPCH, &val);
        r = dsp_cmd(HH_OTHER);
    }
    return(val);
}

int 
get_opch(void)
{
    int r, val=0;

    dsp_get(OPCH, &val);
    return(val);
}

int 
set_sclks(int sclks)    /* A=0x01 B=0x02 both=0x03 */
{
    int r, val=0;

    if (sclks==0x01 || sclks==0x02){
        r = dsp_put(HDIR, sclks);
        if (r == 0)dsp_get(HDIR, &val);
        r = dsp_cmd(HH_OTHER);
    }
    return(val);
}

int 
get_sclks(void)
{
    int r, val=0;

    dsp_get(HDIR, &val);
    return(val);
}

int 
set_pid(int pid)    /* ON=0x1 OFF=0x0 */
{
    int r, val=0;

    if (pid==0x1 || pid==0x0){
        r = dsp_put(PID, pid);
        if (r == 0)dsp_get(PID, &val);
    }
    return(val);
}

int 
get_pid(void)
{
    int r, val=0;

    dsp_get(PID, &val);
    return(val);
}


/*============================================================================*/
int
set_dacs(void)
{
    int r;

    /* Send command to DSP via SSI */
    r = dsp_cmd(HH_DACS);
    return(0);
}

int
base_set_temp(double temp[])
{
}

int
base_get_temp(double temp[])
{
    const int NTEMP=8;
    int r, j, val;

    double s =  1000.0;         /* ADC converter DN/volts */
    double k =  0.385;          /* RTD gain ohms/degC */
    double G =  51.0;           /* in-amp gain = 51 */
    double I0 = 500.0E-6;       /* sensor bias current */
    double R0 =   100.0;        /* sensor nominal ohms at 0 degC */
    double DN = 0.0;

    dsp_clear_fifo();
    dsp_cmd(HH_TEMP);

    udelay(UTIMEOUT);

    for (j=0; j<NTEMP; j++){
        r = dsp_get(TCCD+j, &val);
        if (r!=0)return(-1);
        temp[j] = (((double)val)/(s*G*I0)-R0)/k;
    }

    return(0);
}

int
base_set_tset(double tset[])
{
    const int NTEMP=8;
    int r, j, val;
    double s =  1000.0;         /* ADC converter DN/volts */
    double k =  0.385;          /* RTD gain ohms/degC */
    double G =  51.0;           /* in-amp gain = 51 */
    double I0 = 500.0E-6;       /* sensor bias current */
    double R0 = 100.0;          /* sensor nominal ohms at 0 degC */

    for (j=0; j<NTEMP; j++){
        val = (int)((k*tset[j] + R0)*(s*G*I0));
        r = dsp_put(TSET+j, val);
    }
    return(0);
}


int
base_get_tset(double tset[])
{
    const int NTEMP=8;
    int r, j, val;

    double s =  1000.0;         /* ADC converter DN/volts */
    double k =  0.385;          /* RTD gain ohms/degC */
    double G =  51.0;           /* in-amp gain = 51 */
    double I0 = 500.0E-6;       /* sensor bias current */
    double R0 = 100.0;          /* sensor nominal ohms at 0 degC */

    dsp_clear_fifo();

    for (j=0; j<NTEMP; j++){
        r = dsp_get(TSET+j, &val);
        if (r!=0)return(-1);
        tset[j] = (((double)val)/(s*G*I0)-R0)/k;
    }
    return(0);
}

int
base_set_heat(double heat[])
{
    const int NHEAT=8;
    int j, val[8];
    int r=0;
    double s=20.0;          /* .050 mA/DN  */

    for (j=0; j<NHEAT; j++){
        val[j] = DZ+(int)(s*heat[j]);
        r = dsp_put(HEAT+j, val[j]);
    }
    r = dsp_cmd(HH_TEC);
}

int
base_get_heat(double heat[])
{
    const int NHEAT=8;
    int j, val[8];
    int r=0;
    double s=20.0;          /* .050 mA/DN  */

    for (j=0; j<NHEAT; j++){
        r = dsp_get(HEAT+j, &val[j]);
        heat[j] = (int)((val[j]-DZ)/s);
    }
}

int
base_set_vdac(double vbias[], double vclk[], int ccd)
{
    int r1, r2;
    r1=base_set_vbias(vbias, ccd);
    r2=base_set_vclks(vclk, ccd);

    if (r1==0 && r2==0) set_dacs();

    return(0);
}

int
base_get_vdac(double vbias[], double vclk[], int ccd)
{
    base_get_vbias(vbias, ccd);
    base_get_vclks(vclk, ccd);

    return(0);
}

int
base_set_vbias(double vbias[], int ccd)
{
//  OFFA, ADJA, OGA, ABGA, RDA, SPA, OFDA, ODA,
//  OFFB, ADJB, OGB, ABGB, RDB, SPB, OFDB, ODB
    const int NBIAS=16;
    int j, n, r, val;
    double vmin[]={-10.0, -10.0, -10.0, -10.0, +0.0, +0.0, +0.0, +0.0,
        -10.0, -10.0, -10.0, -10.0, +0.0, +0.0, +0.0, +0.0};
    double vmax[]={+10.0, +10.0, +10.0, +10.0, +18.0, +20.0, +28.0, +28.0,
        +10.0, +10.0, +10.0, +10.0, +18.0, +20.0, +28.0, +28.0};
    double gain[]={LO_G, LO_G, LO_G, LO_G, HI_G, HI_G, HI_G, HI_G,
        LO_G, LO_G, LO_G, LO_G, HI_G, HI_G, HI_G, HI_G};

    n = NBIAS*(int)(ccd/2);
    for (j=0; j<NBIAS; j++){
        if (vbias[j] >= vmin[j] && vbias[j] <= vmax[j]){
            val = DZ+(int)(vbias[j]/gain[j]);
            r = dsp_put(VBIAS+n+j, val);
        }
        else {
            printf("Invalid vbias[%d]  %8.1f \n",j,vbias[j]);
            return(-1);
        }
    }
    return(0);
}

int
base_get_vbias(double vbias[], int ccd)
{
    const int NBIAS=16;
    int j, n, val;
    int r=0;
    double gain[]={LO_G, LO_G, LO_G, LO_G, HI_G, HI_G, HI_G, HI_G,
        LO_G, LO_G, LO_G, LO_G, HI_G, HI_G, HI_G, HI_G};

    n = NBIAS*(int)(ccd/2);
    for (j=0; j<NBIAS; j++){
        r = dsp_get(VBIAS+n+j, &val);
        vbias[j] = (val-DZ)*gain[j];
    }
    return(0);
}

int
base_set_vclks(double vclk[], int ccd)
{
    const int NCLKS=32;
    int j, n, r, val;

// IPCA_L, IPCA_H, IPCB_L, IPCB_H, RGA_L, RGA_H, RGB_L, RGB_H,
// SWA_L, SWA_H, SWB_L, SWB_H, TGA_L, TGA_H, TGB_L, TGB_H,
// SCLKA_L, SCLKA_H, SCLKB_L, SCLKB_H, P1A_L, P1A_H, P2A_L, P2A_H,
// P3A_L, P3A_H, P1B_L, P1B_H, P2B_L, P2B_H, P3B_L, P3B_H

    double vmin[]={-1.0, +0.0, -1.0, +0.0, -10.0, +0.0, -10.0, +0.0,
        -10.0, -2.0, -10.0, -2.0, -12.0, -12.0, -12.0, -12.0,
        -10.0, -10.0, -10.0, -10.0, -12.0, -2.0, -12.0, -2.0,
        -12.0, -2.0, -12.0, -2.0, -12.0, -2.0, -12.0, -2.0};
    double vmax[]={0.0, +6.0, 0.0, +6.0, +10.0, +15.0, +10.0, +15.0,
        +5.0, +10.0, +5.0, +10.0, +5.0, +10.0, +5.0, +10.0,
        +5.0, +10.0, +5.0, +10.0, 0.0, +10.0, 0.0, +10.0,
        0.0, +10.0, 0.0, +10.0, 0.0, +10.0, 0.0, +10.0};
    double gain=LO_G;

    n = NCLKS*(int)(ccd/2);
    for (j=0; j<NCLKS; j++){
        if (vclk[j] >= vmin[j] && vclk[j] <= vmax[j]){
            val = DZ+(int)(vclk[j]/gain);
            r = dsp_put(VCLK+n+j, val);
        }
         else {
            printf("Invalid vclk[%d]  %8.1f \n",j,vclk[j]);
            return(-1);
        }
   }
    return(0);
}


int
base_get_vclks(double vclks[], int ccd)
{
    const int NCLKS=32;
    int j, n, val;
    int r=0;
    double gain=LO_G;

    n = NCLKS*(int)(ccd/2);
    for (j=0; j<NCLKS; j++){
        r = dsp_get(VCLK+n+j, &val);
        vclks[j] = (val-DZ)*gain;
    }
    return(0);
}

int
base_pid(double tset[], double temp[], double heat[], int ccd)
{
#ifdef BASE_TEST
    int e0[]={0,0,0,0,0,0,0,0};
    int h0[]={0,0,0,0,0,0,0,0};

    const int G0=0xD0,G1=0xD1,GE0=0xD2,GE1=0xD3;
    const int E0=0xD8,E1=0xE0,H0=0xE8,H1=0xF0;
    int j;

    for (j=0; j<4; j++){
        dsp_get(E0+j,&e0[j]);
        dsp_get(H0+j,&h0[j]);
    }

    for (j=0; j<2; j++){
        printf("Tccd=%6.1f  Tset=%6.1f  Heat=%6.1f e0=%04x h0=%04x\n",
            temp[j],tset[j],heat[j],e0[j]&0xffff,h0[j]&0xffff);
    }
#endif // BASE_TEST
    return(0);
}


/*============================================================================*/
int
gcam_set_tec(double tec[])
{
    int r, val=0;
    val = DZ + (int)(tec[0]/LO_G);

//    r = dsp_put(TEC, val);        // code from previous version
//    if (r == 0) dsp_get(TEC, &val);

    r = dsp_cmd(HH_TEC);
    dsp_write16(val);
    return(r);
}

int
gcam_get_tec(double tec[])
{
    int r, val=0;

    r = dsp_get(TEC, &val);
    tec[0] = (val - DZ)*LO_G;
    return(r);
}

int
gcam_get_temp(double temp[])
{
    int r, val=0, bias=0;
    double s = 305.176E-6;          /* ADC converter volts/DN */
    double k = -100.0;              /* AD590 gain */
    double G = -2.0;                /* sensor op-amp gain = 0.5 */
    double g = LO_G;                /* bias op-amp gain = 5 mV/DN */
    double V0 =  2.5;               /* ADC min input = -2.5 volts */
    int T0 = -273;

    r = dsp_clear_fifo();
    r = dsp_cmd(HH_TEMP);           // grab temp measurement
    msleep(50);                     // wait for settling + adc

//    r = dsp_get(TEMP, &val);      // code from previous version
//    r = dsp_get(VBIAS, &bias);

    r = dsp_read16(&val);
    r = dsp_read16(&bias);

    /*  T = T0 + k[G(s*ADC - V0) - g(B - DZ)] */
    temp[0] = T0 + (k*(G*(s*val-V0) - g*(bias-DZ)));
    return(0);
}

int
gcam_set_vdac(double vbias[], double vclk[])
{
    int r1, r2;
    r1=gcam_set_vbias(vbias);
    r2=gcam_set_vclks(vclk);

    if (r1==0 && r2==0) set_dacs();

    return(0);
}

int
gcam_get_vdac(double vbias[], double vclk[])
{
    gcam_get_vbias(vbias);
    gcam_get_vclks(vclk);

    return(0);
}

int
gcam_set_vbias(double vbias[])
{
// OFF_R, OFF_L, TEC, OG, SPARE, RD, ODR, ODL

    const int NBIAS=8;
    int j, n, r, val;
    double vmin[]={-10.0, -10.0, -0.0, -10.0, 0.0, 0.0, 0.0, 0.0};
    double vmax[]={+10.0, +10.0, +5.0, +6.0, +25.0, +15.0, +25.0, +25.0};
    double gain[]={LO_G, LO_G, LO_G, LO_G, HI_G, HI_G, HI_G, HI_G};

    for (j=0; j<NBIAS; j++){
        if (vbias[j] >= vmin[j] && vbias[j] <= vmax[j]){
            val = DZ+(int)(vbias[j]/gain[j]);
            r = dsp_put(VBIAS+j, val);
        }
        else {
            printf("Invalid vbias[%d]  %8.1f \n",j,vbias[j]);
            return(-1);
        }
    }
    return(0);
}

int
gcam_get_vbias(double vbias[])
{
    const int NBIAS=8;
    int j, n, val;
    int r=0;
    double gain[]={LO_G, LO_G, LO_G, LO_G, HI_G, HI_G, HI_G, HI_G};

    for (j=0; j<NBIAS; j++){
        n = j%8;
        r = dsp_get(VBIAS+j, &val);
        vbias[j] = (val-DZ)*gain[n];
    }
    return(0);
}

int
gcam_set_vclks(double vclk[])
{
//  IPC_L, IPC_H, RG_L, RG_H, SCLK_L, SCLK_H, DG_L, DG_H
//  TG_L, TG_H, P1_L, P1_H, P2_L, P2_H, P3_L, P3_H
    const int NCLKS=16;
    int j, n, r, val;

    double vmin[]={-1.0, +0.0, -10.0, +0.0, -10.0, +0.0, -10.0, +0.0,
        -10.0, +0.0, -10.0, +0.0, -10.0, +0.0, -10.0, +0.0};
    double vmax[]={+0.0, +6.0, +0.0, +8.0, +0.0, +8.0, +0.0, +8.0,
        +0.0, +8.0, +0.0, +8.0, +0.0, +8.0, +0.0, +8.0};
    double gain=LO_G;

    for (j=0; j<NCLKS; j++){
        if (vclk[j] >= vmin[j] && vclk[j] <= vmax[j]){
            val = DZ+(int)(vclk[j]/gain);
            r = dsp_put(VCLK+j, val);
        }
         else {
            printf("Invalid vclk[%d]  %8.1f \n",j,vclk[j]);
            return(-1);
        }
    }
    return(0);
}

int
gcam_get_vclks(double vclks[])
{
    const int NCLKS=16;
    int j, n, val;
    int r=0;
    double gain=LO_G;

    for (j=0; j<NCLKS; j++){
        r = dsp_get(VCLK+j, &val);
        vclks[j] = (val-DZ)*gain;
    }
    return(0);
}
