/*----------------------------------------------------------------------*
 *  MOTXER.C -- DSP56300 CCD controller software
 *
 *  Original xccdxer version by Ron Johnson  (johnson@astro.ubc.ca)
 *  Updated  motxer  version by Ron Johnson
 *
 *  Updates and modifications by Greg Burley (burley@ociw.edu)
 *
 *----------------------------------------------------------------------*/
#define MOTXER_C

#include <stdio.h>
#include <stddef.h>
#include <stdlib.h>
#include <memory.h>
#include <string.h>
#include <fcntl.h>
#include <math.h>
#include <malloc.h>
#include <sys/types.h>
#include <sys/time.h>
#include "motxer.h"

/*========== Function prototypes ============================================*/
static void set_prgstart_time(void);
static void cx_init(void);
static void cx_exit(void);
static void set_init_defaults(void);
void check_timer(void);

/*========== Global variables ===============================================*/
char tmp_buf[80];
char version[]  = "MOTXER V2.5";
char initrc[]   = ".motxrc";

char cmd1[88]   = { "not defined" };
char cmd2[88]   = { "not defined" };
char cmd3[88]   = { "not defined" };
char cmd4[88]   = { "not defined" };

int dd_state    = 0;        /* open/closeed state of display device */
int dsp_state   = 0;        /* dsp link open=1 not_open=0 */
int expflag     = 0;        /* exposure is in progress */
int prev_state  = 0;        /* prev state of expflag */
int autoshutter = 0;        /* shutter - manual=0 auto=1 special=2 */
int read_err    = 0;        /* 1 if err on xfer reading ccd */
int showflag    = 1;        /* auto show after readout, 1=auto */
int cycleflag   = 0;        /* auto start new exposure, 0=off */
int autosave    = 0;        /* auto file-save after read, 0=off */
int nfsave      = 0;        /* auto filename suffix/num ,for next file */
int autowipe    = 0;        /* auto wipe the array when idle */
int autoflush   = 0;        /* auto-flush ccd before readout */
int autotemp    = 0;        /* auto read temperature sensor */
int pidflag     = 0;        /* temp-control pid loop */
int rdflag      = 0;        /* read line-by-line flag */
int loop_num    = 1;        /* current exposure/loop number */
int hhflag      = 0;        /* hard-hat values need update */
int seqflag     = 0;        /* seq values need update */
int wipeflag    = 0;        /* auto-wipe needs update */
int verbose     = 0;        /* verbose error reporting flag */
double exptime  = 0.0;      /* default exposure time */
double tleft    = 0.0;      /* time left in current timing count */
double cyctime  = 0.0;      /* cycle time (in cycle mode) */
double prgtime  = 0.0;      /* elapsed time since start of program.*/
double endtime  = 0.0;      /* endtime = prgtime + exptime */
double expopen  = 0.0;      /* time shutter was opened */
double expshut  = 0.0;      /* time shutter was closed  */
double systime  = 0.0;      /* current system (gmt) time */
double prgstart = 0.0;      /* unix system time at start of program */
double hhtime   = 0.0;      /* Hardhat update time */

char fprefix[80]  = "ccd";  /* auto filename prefix, user set */
char impath[80]   = "./";   /* image path */
char codepath[80] = "./";   /* code file path */
char fname[80]    = "ccd";  /* image file name */
char *suffix[8] ={"a","b","c","d","e","f","g","h"};
char *channel[8]={"A","B","C","D","E","F","G","H"};

int imsize = 0;             /* size of imbuf (bytes) */
int xbsize = 0;             /* size of extra image buffers */
int xbnum  = 0 ;            /* number of extra buffers */

#ifdef GCAM
const int NCLKS=16;
const int NBIAS=8;
const int NTEMP=8;
const int NHEAT=8;
#endif // GCAM

#ifdef BASE
const int NBIAS=16;
const int NCLKS=32;
const int NTEMP=8;
const int NHEAT=8;
#endif // BASE

u_short *imbuf;             /* allocated in cx_init() */
int *dispbuf;               /* allocated in cx_init() */
IMDESC  image_info, *im = &image_info;
SEQ_MODE seqmode, *seq = &seqmode;

/*========== Main Program ===================================================*/
main(char argc, char* argv[])
{
    const int interval = 10000;

    win_init(argc,argv);            /* Initialize window system */
    cx_init();
    set_init_defaults();
    set_prgstart_time();            /* Set program start time */
    win_start_timer(interval);
    exit(0);
}

/*========== Initialization functions =======================================*/
static void
set_prgstart_time(void)
{
    extern double prgstart;
    struct timeval sys_clock;
    const int sample_time = 1000;
    const int tenth_sec = 100000;

    /* prgstart is the unix time (seconds) at the start of program. */
    /* Exposure timimg is based on the program run time. */
    gettimeofday(&sys_clock, NULL);

    /* wait till next whole tenth of a second */
    while ((sys_clock.tv_usec%tenth_sec) > sample_time) {
        gettimeofday(&sys_clock, NULL);
    }
    prgstart = (sys_clock.tv_sec) + (sys_clock.tv_usec/1.0E6);
}

static void
cx_init(void)
{
    extern u_short *imbuf;
    extern int *dispbuf;
    extern int imsize;
    extern IMDESC *im;
    int r;

    printf("Initializing the CCD sequencer program \n");

    /* get image buffers */
    imbuf = (u_short*)malloc(DEFAULT_IMSIZE*sizeof(short));
    dispbuf = (int*)malloc(DEFAULT_IMSIZE*sizeof(int));
    im->pix[0] = (u_short*)malloc(DEFAULT_IMSIZE*sizeof(short));
    imsize = DEFAULT_IMSIZE;
    xbsize = DEFAULT_IMSIZE;
    xbnum = 1;

    if(imbuf==NULL || dispbuf==NULL || im->pix[0]==NULL) {
        printf("Unable to alloc image buffers\n");
        exit(0);
    }

    /* Initialize the image descriptor info and assign the pix buffer */
    im->pixbuf = im->pix[0];
    imdesc_init(im);

    /* Set the display device configuration */
    dd_state = 0;
    r = 0;
    while (dd_state==0 && r<3){
        dd_state = ddopen();
        if (dd_state)break;
        sleep(1);
        printf("CCD control program requires ximtool -- retrying\n");
        r++;
    }

    if(dd_state) { 
        dderase();
        printf("Display device opened \n");
    }

    /* Open the serial link to the DSP */
    dsp_state = dsp_open();
    if(dsp_state > 0) {
        printf("Serial link opened to DSP \n");
    }
}

static void  
set_init_defaults(void)
{
    extern SEQ_MODE  *seq;
    extern char initrc[];
    extern double exptime;
    extern const int NTEMP,NHEAT,NBIAS,NCLKS;
    char fpath[80];
    int j,r;

    printf("Initializing default parameters \n");
    exptime  = 1.0;

    /* Set the default sequencer parameters */
    seq->mode  = NO_MODE;
    seq->nrow  = DEFAULT_NROW;
    seq->ncol  = DEFAULT_NCOL;
    seq->nft   = DEFAULT_NFT;
    seq->nflush  = DEFAULT_NFLUSH;
    seq->seqtime = DEFAULT_SEQTIME;
    seq->nccd  = MIN_NCCD;
    seq->namps = MIN_NAMPS;
    seq->vbin  = MIN_VBIN;
    seq->hbin  = MIN_HBIN;
    seq->vskip = MIN_VSKIP;
    seq->hskip = MIN_HSKIP;
    seq->vsub  = MIN_VSUB;
    seq->hsub  = MIN_HSUB;
    seq->nshuffle  = MIN_NSHUFFLE;
    seq->gain  = MIN_GAIN;
    seq->usec  = MIN_USEC;
    seq->opch = MIN_OPCH;
    seq->sclks = MIN_SCLKS;
    seq->pid   = MIN_PID;

    seq->npix  = 0;

    for (j=0; j<2; j++)seq->tec[j]=0.0;
    for (j=0; j<NTEMP; j++)seq->tset[j]=0.0;
    for (j=0; j<NTEMP; j++)seq->temp[j]=0.0;
    for (j=0; j<NHEAT; j++)seq->heat[j]=0.0;
    for (j=0; j<NBIAS; j++)seq->vbias[j]=0.0;
    for (j=0; j<NCLKS; j++)seq->vclk[j] =0.0;

    ud_xtime();
    ud_tleft();
    ud_cyctime();
    ud_nfsave();
    ud_exp_state();
    ud_impath();
    ud_cycle(CYCLE_OFF);
    ud_autoshow(ASSERTED);
    ud_autosave(!ASSERTED);
    ud_autowipe(!ASSERTED);

    ud_nrows(seq->nrow);
    ud_ncols(seq->ncol);
    ud_hbin(seq->hbin);
    ud_vbin(seq->vbin);
    ud_gain(seq->gain);
    ud_opch(seq->opch);
    ud_sclks(seq->sclks);
    ud_loop_num(loop_num);
    ud_usec(seq->usec);
    ud_nccd(seq->nccd);

//    ud_temp();            // not necessary?
    ud_infopanel();
    ud_hhpanel();

    /* look for an initrc command file and execute if it exists */
    scriptrun(initrc);
}

static void
cx_exit(void)
{
    extern u_short  *imbuf;

    free(imbuf);
    ddclose();
    dsp_close();
    exit(3);
}

/*===========================================================================*/
void
check_timer(void)
{
    extern IMDESC *im;
    extern int expflag, autoshutter, cycleflag;
    extern int showflag, read_err, autosave, nfsave;
    extern char fprefix[], impath[], fname[];
    extern double xtime, tleft, cyctime, prgtime, endtime;
    extern double expopen, expshut, systime, prgstart;
    static struct timeval sys_clock, *sys=&sys_clock;
    int r;
    char  *bufp, tmp_buf[12];
    extern char  *wx_gotcl;
    const int tenth_sec = 100000;
    extern CMD_ENTRY cmd_dict[];
    extern int dict_size;

    /* First get UT so we have the correct open & shut times */
    /* Then set current program (elapsed) run time */
    gettimeofday(&sys_clock, NULL);
    systime = (sys_clock.tv_sec) + (sys_clock.tv_usec/1.E6);
    prgtime = systime - prgstart;

    /* if a whole second set flag to update header time */
    if(sys_clock.tv_usec <= tenth_sec) ud_clock();

    if (systime > hhtime+10.0){
        if (autotemp)hhflag = GET_TEMP;
//      if (autowipe)wipeflag = SET_WIPE;
        hhtime=systime;
    }

    /* Timing and control logic */
    switch (expflag) {
    case IDLE:                          /* nothing happens */
        if (hhflag != 0) expflag = HHSYNC;
        if (seqflag != 0) expflag = HHSYNC;
        if (wipeflag !=0) expflag = HHSYNC;
        break;
    case CCDFLUSH:                      /* auto-flush the ccd */
        clear_ccd();
        expflag = EXPSTART;
        break;
    case EXPSTART:                      /* start an exposure */
        tleft = exptime;
        endtime = prgtime + exptime;	/* dsp does exposure timing */
        expopen = systime;
        ccd_expose();
        expflag = EXPOSE;
        break;
    case EXPOSE:
        tleft = endtime - prgtime;      /* update time left in exp */
        if(tleft <= 0.0) {              /* end of exposure? */
            expshut = systime;
            expflag = CCDREAD;
            printf("Exposure time  %.1f seconds\n",expshut-expopen);
        }
        ud_tleft();
        break;
    case EXPABORT:
        tleft = 0.0;
        ud_tleft();
        cycleflag = CYCLE_OFF;
        ud_cycle(CYCLE_OFF);
        expflag = IDLE;
        break;
    case CCDREAD:                       /* initiate a readout */
#ifdef BASE
        msleep(2000);
#endif // BASE
        if (autoshutter==2)set_shutter(OPEN_SHUTTER);
        r = ccd_read();
        if (autoshutter==2)set_shutter(CLOSE_SHUTTER);
        if(r == ABORT) {
            cycleflag = CYCLE_OFF;
            ud_cycle(CYCLE_OFF);
            autosave = !ASSERTED;
            ud_autosave(!ASSERTED);
        }
        expflag = FILESAVE;
        break;
    case FILESAVE:
        if(autosave == ASSERTED) {
            save_fitsimage();
            nfsave++;
            ud_nfsave();
        }
        expflag = SHOWNEW;
        break;
    case SHOWNEW:
        if(showflag == ASSERTED) show_image();
        expflag = DELAYSTART;
        break;
    case DELAYSTART:                    /* start delay timer */
        if (cycleflag == CYCLE_ON)endtime=expopen+cyctime;
        if (cycleflag == CYCLE_OFF)endtime=systime;
        loop_num++;
        ud_loop_num(loop_num);
        expflag = EXPDELAY;
        break;
    case EXPDELAY:                      /* delay timer */
        tleft = endtime - systime;
        if(tleft < 0.0) expflag = NEXTCYCLE;
        ud_tleft();
        break;
    case NEXTCYCLE:
        if(cycleflag == CYCLE_ON)   expflag = EXPSTART;
        if(cycleflag == CYCLE_OFF) {
            expflag = IDLE;
            loop_num = 1;
        }
        break;
    case HHSYNC:
        expflag = IDLE;
        if (seqflag == SET_VDAC) hh_set_vdacs();
        if (seqflag == GET_VDAC) hh_get_vdacs();
        if (seqflag == SET_NROWS) hh_set_nrows();
        if (seqflag == SET_NCOLS) hh_set_ncols();
        if (seqflag == SET_HBIN) hh_set_hbin();
        if (seqflag == SET_VBIN) hh_set_vbin();
        if (seqflag == SET_GAIN) hh_set_gain();
        if (seqflag == SET_OPCH) hh_set_opch();
        if (seqflag == SET_SCLKS) hh_set_sclks();
        if (seqflag == SET_XTIME) set_exptimer();
        if (seqflag == SET_CYCTIME) set_cyctime();
        if (seqflag == SET_LOOPNUM) set_loop_num();
        if (seqflag == SET_FSAVE) set_nfsave();
        if (seqflag == SET_USEC) hh_set_usec();
        if (seqflag == SET_NCCD) hh_set_nccd();
        if (seqflag == SET_PID) hh_set_pid();
        if (seqflag == SET_HEAT) hh_set_heat();
        if (hhflag == GET_TEMP) hh_get_ccdtemp();
        if (hhflag == GET_HEAT) hh_get_heat();
        if (hhflag == SET_TEMP) hh_set_ccdtemp();
        if (hhflag == SET_HEAT) hh_set_heat();
//        if (seqflag == SET_WIPE) set_idlewipe();

        if (wipeflag == SET_WIPE) {
            if (autowipe==1)set_idlewipe();
            if (autowipe==0)set_seqmode(NO_MODE);
        }
        wipeflag = IDLE;
        hhflag = IDLE;
        seqflag = SEQ_NONE;
        break;
    default:
        expflag = IDLE;
        break;
    }

    if((bufp = wx_gotcl) != NULL) {         /* got a cmd line */
        wx_gotcl = NULL;
        fp_tokenize(bufp);                  /* parse a command line */
        fp_parse(cmd_dict, dict_size);
    }

    if(prev_state != expflag) {
        ud_exp_state();
        prev_state = expflag;
    }
}


/*==================== End Of File ======================================== */
