/*===========================================================================*/
/* Plot device routines                                                      */
/*===========================================================================*/

#define PLOTDEV_C
#include <stdio.h>
#include <stddef.h>
#include <X11/Xlib.h>
#include <X11/Xutil.h>
#include <X11/cursorfont.h>
#include <X11/keysym.h>

#include <math.h>
#include <sys/stat.h>
#include <sys/types.h>
#include "motxer.h"


/*========== definitions ====================================================*/
#define  SIZE_CMAP          9               /* number of colours */
#define  MINCOLOR           1               /* color zero = background */
#define  MAXCOLOR           (SIZE_CMAP-1)
#define  NSEGS              200             /* size of segment buffer */
#define  SC_MAX             32767
#define  LOC_X              200
#define  LOC_Y              200
#define  WIN_XSIZE          600
#define  WIN_YSIZE          500

/* indices into color table renders specified colors. */
#define  DRKBLU         0
#define  GREEN          1
#define  YELLOW         2
#define  RED            3
#define  CYAN           4
#define  ORANGE         5
#define  MAGENTA        6
#define  VIOLET         7
#define  WHITE          8

/*========== function prototypes ============================================*/
static int init_cursor(void);
static void color_set_up(Display *the_display);
static void reset_colors(Display *display, GC gc);

/*========== local variables ================================================*/
static Display  *px_display;
static int      px_screen;
static Window   px_window;
static Window   px_rootwin;
static GC       px_draw_gc;     /* GC used for drawing */

static Font px_font;
static char *px_ncolors[9] = 
    { 
    "", "green", "gold", "cyan", "magenta", 
    "white", "red", "yellow", "darkorchid"
    };

static u_int px_forecolor, px_backcolor;

static Colormap px_color_map;
static XColor   px_colors[8];
static XColor   px_exact_colors[8];
static Cursor   px_Gcursor;

/*---------------------------------------------------------------------*/

static int px_is_open = 0;              /* initialise frame */
static int px_nseg    = 0;              /* counter in points[] */
static unsigned int px_xsize   = WIN_XSIZE;      /* width  of screen */
static unsigned int px_ysize   = WIN_YSIZE;      /* height of screen */
static int px_xloca   = LOC_X;          /* x location of screen */
static int px_yloca   = LOC_Y;          /* y location of screen */

static int px_aspect = (int)(((double)WIN_YSIZE/(double)WIN_XSIZE) * 1000.0);
static double px_xscale = (double)WIN_XSIZE/(double)SC_MAX;
static double px_yscale = (double)WIN_YSIZE/(double)SC_MAX;
static int px_cursx = -1;               /* last position of cursor */
static int px_cursy = -1;

static XSegment px_seglist[NSEGS];      /* buffer for segment list */


/*========== routines =======================================================*/

/* create a bull's eye cursor and assigns it to the window (parameter) */
static int
init_cursor(void)
{
    px_Gcursor = XCreateFontCursor(px_display, XC_circle);
    XDefineCursor(px_display, px_window, px_Gcursor);
}


/* set up colors */
static void
color_set_up(Display *the_display)
{
    int i;

    px_color_map = DefaultColormap(the_display, 0);

    for (i=1; i<9; i++){
        XAllocNamedColor(the_display, px_color_map, px_ncolors[i],
            &px_exact_colors[i], &px_colors[i]);
    }
}



static void
reset_colors(Display *display, GC gc)
{
    XSetForeground(display, gc, px_forecolor);
}


int
px_init(void)
{
    int i, r, g, b;
    XSizeHints size_hints;

    px_display = XOpenDisplay("");
    px_screen  = DefaultScreen (px_display);
    px_rootwin = RootWindow(px_display, px_screen);

    px_backcolor = BlackPixel(px_display, px_screen);
    px_forecolor = WhitePixel(px_display, px_screen);

    size_hints.x = px_xloca;
    size_hints.y = px_yloca;
    size_hints.width  = px_xsize;
    size_hints.height = px_ysize;
    size_hints.flags = PPosition | PSize;

    px_window = XCreateSimpleWindow(px_display, px_rootwin,
                size_hints.x, size_hints.y, size_hints.width, size_hints.height,
                5, px_forecolor,    /* border width & color */
                px_backcolor);

    XSetStandardProperties(px_display, px_window, "Xccdxer-Plot", "ccd-plot",
                None, NULL, 00, &size_hints);

    px_draw_gc = XCreateGC(px_display, px_window, 0, 0);

    XSetBackground(px_display, px_draw_gc, px_backcolor);
    XSetForeground(px_display, px_draw_gc, px_forecolor);

    XSelectInput(px_display, px_window, ExposureMask);

    init_cursor();
    color_set_up(px_display);

    XMapRaised(px_display, px_window);
    return(1);
}



/* Update sizes, scaling factors, and aspect ratio after a resize, etc. */
void
px_resize(void)
{
    unsigned int depth;
    u_int b_width;
    Window root;
    XSizeHints size_hints;

    XGetGeometry(px_display, px_window, &root, &px_xloca, &px_yloca,
                &px_xsize, &px_ysize, &b_width, &depth);

    px_aspect = (int)(((double)px_ysize/(double)px_xsize) * 1000.0);
    px_xscale = (double)px_xsize/(double)SC_MAX;
    px_yscale = (double)px_ysize/(double)SC_MAX;
    px_cursx = px_cursy = -1;

    /* set default size and position to the new values */
    size_hints.x = px_xloca;
    size_hints.y = px_yloca;
    size_hints.width  = px_xsize;
    size_hints.height = px_ysize;
    size_hints.flags = USPosition | PSize;
    XSetStandardProperties(px_display, px_window, "Xccdxer-Plot", "ccd-plot",
                None, NULL, 00, &size_hints);
}

int
px_open(void)
{
    static int got_a_frame = 0;

    if(got_a_frame == 0){
        got_a_frame = px_init();
        if(got_a_frame <= 0) return(-1);
    }

    if(px_is_open == 0){
        XMapRaised(px_display, px_window);
        px_resize();
        XFlush(px_display);
        px_is_open = 1;
    }

    return(0);
}


int
px_plotmode(void)
{
    if(px_is_open == 0) px_open();
    px_resize();
    return(px_aspect);
}


int
px_gflush(void)
{
    if(px_nseg > 0) {
        XDrawSegments(px_display, px_window, px_draw_gc, px_seglist, px_nseg);
        XFlush(px_display);
        px_nseg = 0;
    }
    return(1);
}


int
px_textmode(void)
{
    px_gflush();
    return(1);
}


int
px_close(void)
{
    XUnmapWindow(px_display, px_window);
    XFlush(px_display);
    px_is_open = 00;

    return(0);
}


int
px_line(int x1, int y1, int x2, int y2)
{
    x1 = x1*px_xscale;
    x2 = x2*px_xscale;
    y1 = y1*px_yscale;
    y2 = y2*px_yscale;
    y1  = px_ysize - y1;
    y2  = px_ysize - y2;

    px_seglist[px_nseg].x1 = (short)x1;
    px_seglist[px_nseg].y1 = (short)y1;
    px_seglist[px_nseg].x2 = (short)x2;
    px_seglist[px_nseg].y2 = (short)y2;
    px_nseg++;

    if(px_nseg >= NSEGS - 1) {
        XDrawSegments(px_display, px_window, px_draw_gc, px_seglist, px_nseg);
        XFlush(px_display);
        px_nseg = 0;
    }
}

int
px_color(int colornum)
{
    px_gflush();

    if (colornum < MINCOLOR)
        colornum = MINCOLOR;
    else
        colornum = colornum%MAXCOLOR;

    XSetForeground(px_display, px_draw_gc, px_colors[colornum].pixel);
    return(colornum);
}


int
px_rdcursor(int *x, int *y)
{
    char buff[5];
    int kx, ky;                     /* cursor position in keyboard wind */
    int rx, ry;                     /* cursor position in root wind */
    u_int button_status;            /* not used */
    Window child_wind;              /* the values of these Windows */
    Window root_wind;               /* are not used */
    XEvent event;

    XQueryPointer(px_display, px_rootwin ,
                &root_wind, &child_wind, &rx, &ry,
                &kx, &ky, &button_status);
    if(px_cursx < 0) {
        px_cursx = px_xsize/2;
        px_cursy = px_ysize/2;
    }
    XWarpPointer(px_display, None, px_window, 0,0,0,0, px_cursx, px_cursy);

    px_Gcursor = XCreateFontCursor(px_display, XC_crosshair);
    XDefineCursor(px_display, px_window, px_Gcursor);

    XSelectInput(px_display, px_window, ButtonPressMask|KeyPressMask);
    while(XCheckTypedEvent(px_display, ButtonPress|KeyPress, &event));
    for(;;) {
        XNextEvent(px_display, &event);
        if(event.type == ButtonPress || event.type == KeyPress) break;
    }
    XSelectInput(px_display, px_window, ExposureMask);

    px_Gcursor = XCreateFontCursor(px_display, XC_circle);
    XDefineCursor(px_display, px_window, px_Gcursor);

    XQueryPointer(px_display, px_window,
                &root_wind, &child_wind, &rx, &ry,
                &px_cursx, &px_cursy, &button_status);
    XWarpPointer(px_display, None, px_rootwin, 0,0,0,0, kx, ky);
    XFlush(px_display);         /* make the pointer move NOW */

    *x = (int)(((double)px_cursx/px_xscale) + 0.5);
    *y = (int)(((double)(px_ysize - px_cursy - 1)/px_yscale) + 0.5);

    if(event.type == ButtonPress){
        switch (event.xbutton.button){
            case Button1:   return('e');
            case Button2:   return('p');
            case Button3:   return('q');
            default:        return('p');
        }
    }

    XLookupString((XKeyEvent *)&event,buff, sizeof(buff), (KeySym *)NULL,
                        (XComposeStatus *)NULL);

    return(buff[0]);
}


int
px_lwidth(void)
{

}


int
px_erase(void)
{
    px_nseg = 0;
    XClearWindow(px_display, px_window);
    XFlush(px_display);
}

